use crate::helper::is_user_admin_right;
use crate::{Context, Error};
use clickhouse_pool::traits::Model;
use database::trivial::Trivial;
use database::trivial_question::TrivialQuestion;
use poise::serenity_prelude;
use poise::serenity_prelude::Channel;
use tracing::{info, instrument, warn};

/// Load a csv file containing trivia questions and answers into the database.
#[instrument(name = "trivia_load", skip(ctx), level = "info", fields(channel = ctx.channel_id().get(), guild = ?ctx.guild_id().unwrap().get()))]
#[poise::command(
    prefix_command,
    slash_command,
    guild_only,
    category = "Trivia",
    check = "is_user_admin_right"
)]
pub async fn load(
    ctx: Context<'_>,
    #[description = "Fichier au format CSV contenant questions, réponses"]
    file: serenity_prelude::Attachment,
    #[description = "Channel of the trivia game, or current"] channel_id: Option<Channel>,
    #[description = "Whether or not the file has an header row"] has_header: Option<bool>,
) -> Result<(), Error> {
    let guild_id = match ctx.guild_id() {
        Some(id) => id.get(),
        None => {
            ctx.say("This command can only be used in a server.")
                .await?;
            return Ok(());
        }
    };
    let channel_id = match channel_id {
        Some(c) => c.id().get(),
        None => {
            ctx.say("No channel specified, using current channel.")
                .await?;
            ctx.channel_id().get()
        }
    };
    let manager_pool = ctx.data().datalake_config.clone();
    // Check if the trivia game already exists in the channel
    let trivia = Trivial::build_select_query(
        Some(&format!(
            "channel_id = {} and guild_id = {}",
            channel_id.clone(),
            guild_id.clone()
        )),
        None,
        None,
    );
    let trivia_exists: Vec<Trivial> = match manager_pool.execute_select_with_retry(&trivia).await {
        Ok(trivia) => trivia,
        Err(e) => {
            ctx.say(format!("Failed to fetch trivia game: {}", e))
                .await?;
            return Ok(());
        }
    };
    if trivia_exists.is_empty() {
        ctx.say(format!("No trivia game exists in this channel.",))
            .await?;
        return Ok(());
    }
    let trivia = trivia_exists.first().unwrap();
    let has_header = has_header.unwrap_or(true);
    let file_content = match file.download().await {
        Ok(content) => content,
        Err(e) => {
            ctx.say(format!("Failed to download file: {}", e)).await?;
            return Ok(());
        }
    };
    // Assuming the file is a CSV, we can parse it here.
    let content = String::from_utf8_lossy(&file_content);
    let content_split = if has_header {
        // Skip the first line if it is a header
        let mut lines = content.lines();
        lines.next(); // Skip header
        lines.collect::<Vec<&str>>().into_iter()
    } else {
        content.lines().collect::<Vec<&str>>().into_iter()
    };
    // Split each line by comma and process

    let parsed_content = content_split
        .map(|line| {
            let parts: Vec<&str> = line.split(',').collect();
            if parts.len() < 2 || parts[0].trim().is_empty() || parts[1].trim().is_empty() {
                warn!("Invalid line format: {}", line);
                return ("".to_string(), "".to_string());
            }
            let question = parts[0].trim().to_string();
            let answer = parts[1].trim().to_string();
            (question, answer)
        })
        .filter(|(q, a)| !q.is_empty() && !a.is_empty())
        .collect::<Vec<(String, String)>>();
    if parsed_content.is_empty() {
        ctx.say("No valid questions found in the file.").await?;
        return Ok(());
    }

    let get_trivia = TrivialQuestion::build_select_query(
        Some(&format!(
            "channel_id = {} and guild_id = {}",
            channel_id, guild_id
        )),
        None,
        None,
    );
    let trivial_questions: Vec<TrivialQuestion> =
        match manager_pool.execute_select_with_retry(&get_trivia).await {
            Ok(questions) => questions,
            Err(e) => {
                ctx.say(format!("Failed to fetch trivia questions: {}", e))
                    .await?;
                return Ok(());
            }
        };
    let parsed_content = if !trivial_questions.is_empty() {
        // Check for duplicate questions
        let existing_questions: Vec<String> = trivial_questions
            .iter()
            .map(|q| q.question.clone())
            .collect();
        let no_duplicates: Vec<(String, String)> = parsed_content
            .iter()
            .filter(|(q, _)| !existing_questions.contains(q))
            .cloned()
            .collect();
        if no_duplicates.is_empty() {
            ctx.say("Questions already exist in the database, no new questions to add.")
                .await?;
            return Ok(());
        }
        no_duplicates
    } else {
        // No existing questions, all are new
        parsed_content
    };
    let mut inserter = match manager_pool
        .get_insert::<TrivialQuestion>(TrivialQuestion::table_name())
        .await
    {
        Ok(inserter) => inserter,
        Err(e) => {
            ctx.say(format!(
                "Internal Server error, please contact Bot Developer: {}",
                e
            ))
            .await?;
            warn!("Failed to create inserter for TrivialQuestion: {}", e);
            return Ok(());
        }
    };
    for (question, answer) in parsed_content {
        let trivial_question = TrivialQuestion {
            trivial_id: trivia.id,
            guild_id: guild_id,
            channel_id: channel_id,
            question: question,
            answer: answer,
            creator_id: ctx.author().id.get(),
            updater_id: ctx.author().id.get(),
            ..Default::default()
        };
        if let Err(e) = inserter.write(&trivial_question).await {
            warn!("Failed to insert trivial question: {}", e);
            ctx.say(format!("Failed to insert question: {}", e)).await?;
            return Ok(());
        }
    }
    match inserter.end().await {
        Ok(_) => {
            info!("Successfully loaded trivia questions into the database.");
            ctx.say("Trivia questions loaded successfully.").await?;
        }
        Err(e) => {
            warn!("Failed to finalize insert operation: {}", e);
            ctx.say(format!("Failed to load trivia questions: {}", e))
                .await?;
            return Ok(());
        }
    }

    let optimize = TrivialQuestion::optimize_table_sql();
    if let Err(e) = manager_pool.execute_with_retry(optimize).await {
        warn!("Failed to optimize trivial_question table: {}", e);
        ctx.say(format!("Failed to optimize trivia questions: {}", e))
            .await?;
        return Ok(());
    }
    info!("Successfully optimized trivial_question table.");
    ctx.say("Trivia questions loaded and table optimized successfully.")
        .await?;
    Ok(())
}
