use poise::serenity_prelude::Http;
use std::{
    collections::HashMap,
    fmt::{self, Display},
    sync::Arc,
};
use tokio::sync::RwLock;
use tokio_cron_scheduler::JobScheduler;
use tracing::{error, info, instrument};
use uuid::Uuid;

#[derive(Debug)]
pub enum StopScheduleJob {
    JobNotFound,
    RemoveFailed,
}

impl Display for StopScheduleJob {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{:?}", self)
    }
}

pub struct ScheduleJob {
    pub job_id: Arc<RwLock<HashMap<(u64, u64), Uuid>>>,
    pub scheduler: JobScheduler,
}

impl Clone for ScheduleJob {
    fn clone(&self) -> Self {
        ScheduleJob {
            job_id: self.job_id.clone(),
            scheduler: self.scheduler.clone(),
        }
    }
}

impl ScheduleJob {
    #[instrument(level = "info")]
    pub async fn start_cron_scheduler() -> Result<Self, ()> {
        let scheduler = JobScheduler::new().await;
        let mut future_self = match scheduler {
            Ok(scheduler) => ScheduleJob {
                job_id: Default::default(),
                scheduler,
            },
            Err(e) => {
                error!("Error starting cron scheduler: {:?}", e);
                return Err(());
            }
        };
        future_self.scheduler.set_shutdown_handler(Box::new(|| {
            Box::pin(async {
                info!("Cron scheduler stopped");
            })
        }));
        future_self.scheduler.shutdown_on_ctrl_c();
        match future_self.scheduler.start().await {
            Ok(_) => {
                info!("Cron scheduler started");
                Ok(future_self)
            }
            Err(e) => {
                error!("Error starting cron scheduler: {:?}", e);
                Err(())
            }
        }
    }

    #[instrument(skip(self, _http), level = "info")]
    pub async fn load_all_trivial_cron_job(&mut self, _http: &Http) -> Result<(), bool> {
        // Load all trivial jobs
        Ok(())
    }

    #[instrument(skip(self), level = "info")]
    pub async fn stop_cron_scheduler(&mut self) -> &mut Self {
        let job_id = self.job_id.write().await;

        for (server_id, channel_id) in job_id.keys() {
            match self
                .scheduler
                .remove(job_id.get(&(*server_id, *channel_id)).unwrap())
                .await
            {
                Ok(_) => {
                    info!("Cron job removed");
                }
                Err(e) => {
                    error!("Error removing cron job: {:?}", e);
                }
            }
        }

        drop(job_id);
        match self.scheduler.shutdown().await {
            Ok(_) => {
                info!("Cron scheduler stopped");
                self
            }
            Err(e) => {
                error!("Error stopping cron scheduler: {:?}", e);
                self
            }
        }
    }

    #[instrument(skip(self, http), level = "info")]
    pub async fn add_trivial_cron_job(
        &mut self,
        server_id: u64,
        channel_id: u64,
        cron_expression: String,
        http: &Http,
    ) -> Result<Uuid, ()> {
        let _http = Arc::new(Http::new(http.token()));
        // Create a new job with the provided cron expression
        Err(())
    }
    #[instrument(skip(self), level = "info")]
    pub async fn stop_scheduled_job(
        &mut self,
        server_id: u64,
        channel_id: u64,
    ) -> Result<(), StopScheduleJob> {
        let remove_job = {
            let mut job_id = self.job_id.write().await;
            job_id.remove(&(server_id, channel_id))
        };
        match remove_job {
            Some(job_uid) => match self.scheduler.remove(&job_uid).await {
                Ok(_) => {
                    info!("Cron job removed");
                    Ok(())
                }
                Err(e) => {
                    error!("Error removing cron job: {:?}", e);
                    Err(StopScheduleJob::RemoveFailed)
                }
            },
            None => {
                error!("Cron job not found");
                Err(StopScheduleJob::JobNotFound)
            }
        }
    }
}
