use std::collections::HashMap;

use chrono::{DateTime, Utc};
use clickhouse::Row;
use clickhouse_pool::traits::Model;
use serde::{Deserialize, Serialize};

#[derive(Debug, Clone, PartialEq, Row, Serialize, Deserialize)]
pub struct Guild {
    pub id: u64,
    pub bot_name: String,
    pub enrolled: bool,
    pub prefix: String,
    pub moderator: Vec<u64>,

    pub feature_flags: HashMap<String, bool>,

    #[serde(with = "clickhouse::serde::chrono::datetime64::millis")]
    pub created_at: DateTime<Utc>,
    #[serde(with = "clickhouse::serde::chrono::datetime64::millis")]
    pub updated_at: DateTime<Utc>,
    pub creator_id: u64,
    pub updater_id: u64,
}

impl Model for Guild {
    type T = Guild;

    fn table_name() -> &'static str {
        "guild"
    }

    fn optimize_table_sql() -> &'static str {
        r#"
        OPTIMIZE TABLE guild FINAL
        "#
        .trim()
    }

    fn create_table_sql() -> &'static str {
        r#"
        CREATE TABLE IF NOT EXISTS guild (
            id UInt64 PRIMARY KEY,
            bot_name String,
            enrolled Bool,
            prefix String,
            moderator Array(UInt64),
            feature_flags Map(String, Bool),
            created_at DateTime64(3),
            updated_at DateTime64(3),
            creator_id UInt64,
            updater_id UInt64
        ) ENGINE = MergeTree()
        ORDER BY (id)
        "#
        .trim()
    }

    fn column_names() -> Vec<&'static str> {
        vec![
            "id",
            "bot_name",
            "enrolled",
            "prefix",
            "moderator",
            "feature_flags",
            "created_at",
            "updated_at",
            "creator_id",
            "updater_id",
        ]
    }

    fn to_row(&self) -> (Vec<&'static str>, Vec<String>) {
        (
            Self::column_names(),
            vec![
                self.id.to_string(),
                self.bot_name.clone(),
                self.enrolled.to_string(),
                self.prefix.clone(),
                format!("{:?}", self.moderator),
                format!("{:?}", self.feature_flags),
                self.created_at.to_rfc3339(),
                self.updated_at.to_rfc3339(),
                self.creator_id.to_string(),
                self.updater_id.to_string(),
            ],
        )
    }

    fn insert_query(&self) -> String {
        let (columns, values) = self.to_row();
        let columns_str = columns.join(", ");
        let values_str = values.join(", ");

        format!(
            "INSERT INTO {} ({}) VALUES ({})",
            Self::table_name(),
            columns_str,
            values_str
        )
    }

    fn batch_insert_query(items: &[Self::T]) -> String {
        let mut queries = Vec::new();
        for item in items {
            let (columns, values) = item.to_row();
            let columns_str = columns.join(", ");
            let values_str = values.join(", ");
            queries.push(format!(
                "INSERT INTO {} ({}) VALUES ({})",
                Self::table_name(),
                columns_str,
                values_str
            ));
        }
        queries.join("; ")
    }

    fn build_select_query(
        where_clause: Option<&str>,
        limit: Option<u64>,
        offset: Option<u64>,
    ) -> String {
        let mut query = format!(
            "SELECT {} FROM {}",
            Self::column_names().join(", "),
            Self::table_name()
        );
        if let Some(where_clause) = where_clause {
            query.push_str(&format!(" WHERE {}", where_clause));
        }
        if let Some(limit) = limit {
            query.push_str(&format!(" LIMIT {}", limit));
        }
        if let Some(offset) = offset {
            query.push_str(&format!(" OFFSET {}", offset));
        }
        query
    }
}
